% eda??_??
% simple least squares fit of gaussian with a single tower

% grid of points in (x,y)
Nx = 21;
Ny = 21;
xmax = 5;
ymax = 5;
xmin = 0;
ymin = 0;
x = xmin + (xmax-xmin)* [0:Nx-1]'/(Nx-1);
y = ymin + (ymax-ymin)* [0:Ny-1]'/(Ny-1);

% (x,y) index arrays for easy access
kofij=zeros(Nx,Ny);  % linear index k of a particular (i,j)
iofk=zeros(Nx*Ny,1); % row index of a particular k
jofk=zeros(Nx*Ny,1); % col index of a particular k
Nxy=0;
for i=[1:Nx]
for j=[1:Ny]
    Nxy=Nxy+1;
    kofij(i,j)=Nxy;
    iofk(Nxy)=i;
    jofk(Nxy)=j;
end
end

% Gaussian
dobs = zeros(Nx,Ny);
dobsvec = zeros(Nxy,1);
xbar = 2.4;
ybar = 2.6;
amp = 5.0;
sigmax = 1.2;
sigmay = 0.8;
cx = 1/(2*sigmax*sigmax);
cy = 1/(2*sigmay*sigmay);
for i=[1:Nx]
for j=[1:Ny]
    dobs(i,j) = amp*exp(-cx*((x(i)-xbar)^2) - cy*((y(j)-ybar)^2) );
    dobsvec(kofij(i,j))=dobs(i,j);
end
end

% build the test network for a single 2D tower
Nc = 1;     % One tower
Xc = [2.5]; % Xc: the center of the tower
Yc = [2.5];
Dx = [2];   % Dx: the width of the tower
Dy = [2];
h = [1];    % h: the height of the tower
slope = 5/4;  % slope: slope of the step function

% ---Neural Net definitions---

% Nmax: number of layers;
Lmax = 4;

% Lmax: maximum number of neurons in any layer
Nmax = Nc * 4;

% N(i): column-vector with number of neurons in each layer;
N = zeros(1,Lmax);

% biases: b(1:N(i),i) is a column-vector that gives the
% biases of the N(i) neurons on the i-th layer
b = zeros( Nmax, Lmax );

% weights: w(1:N(i),1:N(i-1),i) is a matrix that gives the
% weights between the N(i) neurons in the i-th layer and the
% N(i-1) neurons on the (i-1)-th layer. The weights assigned
% to the first layer are never used and should be set to zero
w = zeros( Nmax, Nmax, Lmax );

% activity: a(1:N(i),i) is a column-vector that gives the
% activitites of the N(i) neurons on the i-th layer
a = zeros( Nmax, Lmax );

% z: a(1:N(i),i) is a column-vector that gives the
% argument of the activitites of the N(i) neurons on the i-th layer
z = zeros( Nmax, Lmax );

% dada(i,j,k): the change in the activity of the i-th neuron of the layer k
% due to a change in the activity of the j-th neuron in the layer k-1
dada = zeros(Nmax,Nmax,Lmax);

% daLmaxda(i,j,k): the change in the activity of the i-th neuron in layer Lmax
% due to a change in the activity of the j-th neuron in the layer k-1
daLmaxda = zeros(Nmax,Nmax,Lmax);

% dadz(i,k): the change in the activity of the i-th neuron of layer k
% due to a change in the input z of the same neuron
dadz = zeros(Nmax,Lmax);

% daLmaxdb(i,j,k): the change in the activity of the i-th neuron in layer Lmax
% due to a change in the bias of the of the j-th neuron in the layer k
daLmaxdb = zeros(Nmax,Nmax,Lmax);

% daLmaxdw(i,j,k,l): the change in the activity of the i-th neuron in layer Lmax
% due to a change in the weight connecting the j-th neuron in the layer l
% with the k-th neuron of layer l-1
daLmaxdw = zeros(Nmax,Nmax,Nmax,Lmax);

[N,w,b] = eda_net_2dtower( Nc, slope, Xc, Yc, Dx, Dy, h, N, w, b );

figure(2);
clf;
eda_net_view(N,w,b);

% build look-up tables to allow sequential ordering
% of non-trivual biases and weights
layer_of_bias = zeros(Lmax*Nmax,1);
neuron_of_bias = zeros(Lmax*Nmax,1);
index_of_bias = zeros(Nmax,Lmax);
Nb=0;  % number of nontrivial biases
for k=[2:Lmax]
    for j=[1:N(k)]
        Nb = Nb+1;
        layer_of_bias(Nb)=k;
        neuron_of_bias(Nb)=j;
        index_of_bias(j,k)=Nb;
    end
end
Nw=0;  % number of nontrivial weights
layer_of_weight = zeros(Lmax*Nmax*Nmax,1);
r_neuron_of_weight = zeros(Lmax*Nmax*Nmax,1);
l_neuron_of_weight = zeros(Lmax*Nmax*Nmax,1);
index_of_weight = zeros(Nmax,Nmax,Lmax);
for k=[2:Lmax]
    for i=[1:N(k)]
    for j=[1:N(k-1)]
        if( w(i,j,k) ~= 0 )
            Nw = Nw+1;
            layer_of_weight(Nw)=k;
            r_neuron_of_weight(Nw)=i;
            l_neuron_of_weight(Nw)=j;
            index_of_weight(j,i,k)=Nw;
        end
    end
    end
end

% linearized data kernel
M = Nb+Nw;
G = zeros(Nxy,M);

% predicted data
dpre = zeros(Nx,Ny);
dprevec = zeros(Nxy,1);

% initial predicted data
d0 = zeros(Nx,Ny);
d0vec = zeros(Nxy,1);

Niter=20;
for iter=[1:Niter]

% loop over (x,y)
for kx=[1:Nx]
for ky=[1:Ny]
    
    % activities
    a(1:N(1),1) = [x(kx);x(ky)];
    for i=[2:Lmax]
        z(1:N(i),i) = w(1:N(i),1:N(i-1),i) * a(1:N(i-1),i-1) + b(1:N(i),i);
        if i < Lmax
            a(1:N(i),i) = 1.0./(1.0+exp(-z(1:N(i),i)));
        else
            a(1:N(i),i) = z(1:N(i),i);
        end
    end
    dpre(kx,ky) = a(1,4);
    dprevec(kofij(kx,ky))=dpre(kx,ky);

    % da/dz
    for i = [1:Lmax]
        if (i==Lmax)
            % sigma function is sigma(z)=z for top layer
            % d sigma / dz = 1
            dadz(1:N(i),i) = ones(1:N(i),1);
        else
            % sigma function is sigma(z)=(1+exp(-z))^(-1) for other layers
            % d sigma / dz = (-1)(-1) exp(-z)(1+exp(-z))^(-1)
            %              = exp(-z) (a^2)
            % and note exp(-z) = (1/a)-1
            dadz(1:N(i),i) = (a(1:N(i),i).^2).*((1./a(1:N(i),i))-1);
        end
    end
        
    % da/da
    for i = [Lmax:-1:2]
        if i == Lmax
            % a=z in the top of the layer, and
            % z = (sum) w a + b
            % so da(layer i)/da(layer i-1) = dz/da(layer i-1) = w
            dada(1:N(i),1:N(i-1),i) = w(1:N(i),1:N(i-1),i);
            % initialize daLmaxda to prepare for backpropagation
            daLmaxda(1:N(i),1:N(i-1),i) = dada(1:N(i),1:N(i-1),i);
        else
            % a=sigma(z) in the other layers, so use the chain rule
            % da/da = da(layer)/dz * dz/da(layer i-1)
            %    where as before dz/da(layer i-1) = w
            dada(1:N(i),1:N(i-1),i) = diag(dadz(1:N(i),i))*w(1:N(i),1:N(i-1),i);
            % backpropagate
            daLmaxda(1:N(Lmax),1:N(i-1),i) =  ...
                daLmaxda(1:N(Lmax),1:N(i),i+1)*dada(1:N(i),1:N(i-1),i);
        end            
    end
        
    % da/db
    for i = [1:Lmax]
        if( i==Lmax )
            % a=z in the top layer, and z(layer i)= w *a(layer i-1) + b
            % so da(layer i)/db = dz(layer i)/db = 1
            daLmaxdb(1:N(Lmax),1:N(Lmax),Lmax) = eye(N(Lmax),N(Lmax));
        else
            % a = sigma(z) in the other layers, so use the chain rule
            % da(layer i)/db = da(layer i)/dz * dz/db
            % where as before dzdb=1
            dzdb = 1;
            dadb = dadz(1:N(i),i) * dzdb;
            % then apply the chain rule for activities
            % da(layer Lmax)/db = da(layer Lmax)/da(layer i) * da(layer i)/db
            daLmaxdb(1:N(Lmax),1:N(i),i) = daLmaxda(1:N(Lmax),1:N(i),i+1) * diag(dadb);
        end
    end

    % da/dw
    % calculation of weight derivatives
    % since z(layer i) = w*a(layer i-1) + b
    % dz(layer i)/dw = a(layer i-1)
    for l = [2:Lmax]  % for simplicity, loop over right neuron
        if( l==Lmax )
            % in the top layer, a=z, so da/dw = dz/dw = a
            for j = [1:N(l)]
                daLmaxdw(j,j,1:N(l-1),l) = a(1:N(l-1),l-1);
            end
        else
            % in the other layers, use chain rule
            % da/dw = da/dz * dz/dw
            % and then the chain ruke again
            % da(layer Lmax)/dw = da(layer Lmax)/da(layer i) * da(layer i)/dw
            for j = [1:N(l)]
                dzdw = a(1:N(l-1),l-1);
                dadw = dadz(j,l) * dzdw;
                daLmaxdw(1:N(Lmax),j,1:N(l-1),l) = daLmaxda(1:N(Lmax),j,l+1) * dadw;
            end
        end
    end
    
    % biases first
    for ib=[1:Nb]
        nob=neuron_of_bias(ib);
        lob=layer_of_bias(ib);
        G(kofij(kx,ky), ib ) = daLmaxdb(1, nob, lob );
    end
    % weights second
    for iw=[1:Nw]
        rnow = r_neuron_of_weight(iw);
        lnow = l_neuron_of_weight(iw);
        low = layer_of_weight(iw);
        G(kofij(kx,ky), Nb+iw ) = daLmaxdw( 1, rnow, lnow, low );
    end
    
end % next (x,y)
end

Dd = dobsvec - dprevec;
 
if( iter==1 )
    d0 = dpre;
    d0vec = dobsvec;
    error = Dd'*Dd;
    fprintf('starting error %.2f\n', error );
end

% implement some damping to improve stability
% but relax it after a few iterations
if( iter<5 )
    epsi=1.0;
elseif (iter<10)
    epsi=0.1;
else
    epsi=0.001;
end
Dm = (G'*G+epsi*eye(M,M))\(G'*Dd);

% add in bias perturbations
for ib=[1:Nb]
    nob=neuron_of_bias(ib);
    lob=layer_of_bias(ib);
    b(nob, lob) = b(nob, lob) + Dm(ib);
end
% add in weights perturbations
for iw=[1:Nw]
    rnow = r_neuron_of_weight(iw);
    lnow = l_neuron_of_weight(iw);
    low = layer_of_weight(iw);
    w(rnow, lnow, low) = w(rnow, lnow, low) + Dm( Nb+iw );
end

end % next iteration
    
% loop over (x,y)
for kx=[1:Nx]
for ky=[1:Ny]
    
    % activities
    a(1:N(1),1) = [x(kx);x(ky)];
    for i=[2:Lmax]
        z(1:N(i),i) = w(1:N(i),1:N(i-1),i) * a(1:N(i-1),i-1) + b(1:N(i),i);
        if i < Lmax
            a(1:N(i),i) = 1.0./(1.0+exp(-z(1:N(i),i)));
        else
            a(1:N(i),i) = z(1:N(i),i);
        end
    end
    dpre(kx,ky) = a(1,4);
    dprevec(kofij(kx,ky))=dpre(kx,ky);
end
end

eda_draw( dobs, 'caption d-obs', d0, 'caption d-start',dpre, 'caption d-pre');

Dd = dobsvec - dprevec;
error = Dd'*Dd;
fprintf('ending error %.2f\n', error );

